import keras

from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.backbone import Backbone


@keras_hub_export("keras_hub.models.SAMBackbone")
class SAMBackbone(Backbone):
    """A backbone for the Segment Anything Model (SAM).

    Args:
        image_encoder: `keras_hub.models.ViTDetBackbone`. A feature extractor for
            the input images.
        prompt_encoder: `keras_hub.layers.SAMPromptEncoder`. A Keras layer to
            compute embeddings for points, box, and mask prompt.
        mask_decoder: `keras_hub.layers.SAMMaskDecoder`. A Keras layer to
            generate segmentation masks given the embeddings generated by the
            backbone and the prompt encoder.
        dtype: The dtype of the layer weights.

    Example:
    ```python
    image_size=128
    batch_size=2
    input_data = {
        "images": np.ones(
            (batch_size, image_size, image_size, 3),
            dtype="float32",
        ),
        "points": np.ones((batch_size, 1, 2), dtype="float32"),
        "labels": np.ones((batch_size, 1), dtype="float32"),
        "boxes": np.ones((batch_size, 1, 2, 2), dtype="float32"),
        "masks": np.zeros(
            (batch_size, 0, image_size, image_size, 1)
        ),
    }
    image_encoder = keras_hub.models.ViTDetBackbone(
        hidden_size=16,
        num_layers=16,
        intermediate_dim=16 * 4,
        num_heads=16,
        global_attention_layer_indices=[2, 5, 8, 11],
        patch_size=16,
        num_output_channels=8,
        window_size=2,
        image_shape=(image_size, image_size, 3),
    )
    prompt_encoder = keras_hub.layers.SAMPromptEncoder(
        hidden_size=8,
        image_embedding_size=(8, 8),
        input_image_size=(
            image_size,
            image_size,
        ),
        mask_in_channels=16,
    )
    mask_decoder = keras_hub.layers.SAMMaskDecoder(
        num_layers=2,
        hidden_size=8,
        intermediate_dim=32,
        num_heads=8,
        embedding_dim=8,
        num_multimask_outputs=3,
        iou_head_depth=3,
        iou_head_hidden_dim=8,
    )
    backbone = keras_hub.models.SAMBackbone(
        image_encoder=image_encoder,
        prompt_encoder=prompt_encoder,
        mask_decoder=mask_decoder,
        image_shape=(image_size, image_size, 3),
    )
    backbone(input_data)
    ```
    """

    def __init__(
        self,
        image_encoder,
        prompt_encoder,
        mask_decoder,
        dtype=None,
        **kwargs,
    ):
        # === Layers ===
        self.image_encoder = image_encoder
        self.prompt_encoder = prompt_encoder
        self.mask_decoder = mask_decoder
        # === Functional model
        image_input = self.image_encoder.input

        inputs = {
            "images": image_input,
            "points": keras.Input(shape=[None, 2], name="points"),
            "labels": keras.Input(shape=[None], name="labels"),
            "boxes": keras.Input(shape=[None, 2, 2], name="boxes"),
            "masks": keras.Input(shape=[None, None, None, 1], name="masks"),
        }
        image_embeddings = self.image_encoder.output
        prompt_embeddings = self.prompt_encoder(**inputs)
        outputs = {
            "image_embeddings": image_embeddings,
        }
        outputs.update(prompt_embeddings)
        super().__init__(
            inputs=inputs,
            outputs=outputs,
            dtype=dtype,
            **kwargs,
        )

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "image_encoder": keras.layers.serialize(self.image_encoder),
                "prompt_encoder": keras.layers.serialize(self.prompt_encoder),
                "mask_decoder": keras.layers.serialize(self.mask_decoder),
            }
        )
        return config

    @classmethod
    def from_config(cls, config):
        config.update(
            {
                "image_encoder": keras.layers.deserialize(
                    config["image_encoder"]
                ),
                "prompt_encoder": keras.layers.deserialize(
                    config["prompt_encoder"]
                ),
                "mask_decoder": keras.layers.deserialize(
                    config["mask_decoder"]
                ),
            }
        )

        return super().from_config(config)



import keras
from keras import ops


class SpatialPyramidPooling(keras.layers.Layer):
    """Implements the Atrous Spatial Pyramid Pooling.

    Reference for Atrous Spatial Pyramid Pooling [Rethinking Atrous Convolution
    for Semantic Image Segmentation](https://arxiv.org/pdf/1706.05587.pdf) and
    [Encoder-Decoder with Atrous Separable Convolution for Semantic Image
    Segmentation](https://arxiv.org/pdf/1802.02611.pdf)

    Args:
    dilation_rates: list of ints. The dilation rate for parallel dilated conv.
        Usually a sample choice of rates are `[6, 12, 18]`.
    num_channels: int. The number of output channels, defaults to `256`.
    activation: str. Activation to be used, defaults to `relu`.
    dropout: float. The dropout rate of the final projection output after the
        activations and batch norm, defaults to `0.0`, which means no dropout is
        applied to the output.

    Example:
    ```python
    inp = keras.layers.Input((384, 384, 3))
    backbone = keras.applications.EfficientNetB0(
        input_tensor=inp,
        include_top=False)
    output = backbone(inp)
    output = SpatialPyramidPooling(
        dilation_rates=[6, 12, 18])(output)
    ```
    """

    def __init__(
        self,
        dilation_rates,
        num_channels=256,
        activation="relu",
        dropout=0.0,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.dilation_rates = dilation_rates
        self.num_channels = num_channels
        self.activation = activation
        self.dropout = dropout
        self.data_format = keras.config.image_data_format()
        self.channel_axis = -1 if self.data_format == "channels_last" else 1

    def build(self, input_shape):
        channels = input_shape[self.channel_axis]

        # This is the parallel networks that process the input features with
        # different dilation rates. The output from each channel will be merged
        # together and feed to the output.
        self.aspp_parallel_channels = []

        # Channel1 with Conv2D and 1x1 kernel size.
        conv_sequential = keras.Sequential(
            [
                keras.layers.Conv2D(
                    filters=self.num_channels,
                    kernel_size=(1, 1),
                    use_bias=False,
                    data_format=self.data_format,
                    name="aspp_conv_1",
                ),
                keras.layers.BatchNormalization(
                    axis=self.channel_axis, name="aspp_bn_1"
                ),
                keras.layers.Activation(
                    self.activation, name="aspp_activation_1"
                ),
            ]
        )
        conv_sequential.build(input_shape)
        self.aspp_parallel_channels.append(conv_sequential)

        # Channel 2 and afterwards are based on self.dilation_rates, and each of
        # them will have conv2D with 3x3 kernel size.
        for i, dilation_rate in enumerate(self.dilation_rates):
            conv_sequential = keras.Sequential(
                [
                    keras.layers.Conv2D(
                        filters=self.num_channels,
                        kernel_size=(3, 3),
                        padding="same",
                        dilation_rate=dilation_rate,
                        use_bias=False,
                        data_format=self.data_format,
                        name=f"aspp_conv_{i+2}",
                    ),
                    keras.layers.BatchNormalization(
                        axis=self.channel_axis, name=f"aspp_bn_{i+2}"
                    ),
                    keras.layers.Activation(
                        self.activation, name=f"aspp_activation_{i+2}"
                    ),
                ]
            )
            conv_sequential.build(input_shape)
            self.aspp_parallel_channels.append(conv_sequential)

        # Last channel is the global average pooling with conv2D 1x1 kernel.
        if self.channel_axis == -1:
            reshape = keras.layers.Reshape((1, 1, channels), name="reshape")
        else:
            reshape = keras.layers.Reshape((channels, 1, 1), name="reshape")
        pool_sequential = keras.Sequential(
            [
                keras.layers.GlobalAveragePooling2D(
                    data_format=self.data_format, name="average_pooling"
                ),
                reshape,
                keras.layers.Conv2D(
                    filters=self.num_channels,
                    kernel_size=(1, 1),
                    use_bias=False,
                    data_format=self.data_format,
                    name="conv_pooling",
                ),
                keras.layers.BatchNormalization(
                    axis=self.channel_axis, name="bn_pooling"
                ),
                keras.layers.Activation(
                    self.activation, name="activation_pooling"
                ),
            ]
        )
        pool_sequential.build(input_shape)
        self.aspp_parallel_channels.append(pool_sequential)

        # Final projection layers
        projection = keras.Sequential(
            [
                keras.layers.Conv2D(
                    filters=self.num_channels,
                    kernel_size=(1, 1),
                    use_bias=False,
                    data_format=self.data_format,
                    name="conv_projection",
                ),
                keras.layers.BatchNormalization(
                    axis=self.channel_axis, name="bn_projection"
                ),
                keras.layers.Activation(
                    self.activation, name="activation_projection"
                ),
                keras.layers.Dropout(rate=self.dropout, name="dropout"),
            ],
        )
        projection_input_channels = (
            2 + len(self.dilation_rates)
        ) * self.num_channels
        if self.data_format == "channels_first":
            projection.build(
                (input_shape[0],)
                + (projection_input_channels,)
                + (input_shape[2:])
            )
        else:
            projection.build((input_shape[:-1]) + (projection_input_channels,))
        self.projection = projection
        self.built = True

    def call(self, inputs):
        """Calls the Atrous Spatial Pyramid Pooling layer on an input.

        Args:
            inputs: A tensor of shape [batch, height, width, channels]

        Returns:
            A tensor of shape [batch, height, width, num_channels]
        """
        result = []

        for channel in self.aspp_parallel_channels:
            temp = ops.cast(channel(inputs), inputs.dtype)
            result.append(temp)

        image_shape = ops.shape(inputs)
        if self.channel_axis == -1:
            height, width = image_shape[1], image_shape[2]
        else:
            height, width = image_shape[2], image_shape[3]
        result[-1] = keras.layers.Resizing(
            height,
            width,
            interpolation="bilinear",
            data_format=self.data_format,
            name="resizing",
        )(result[-1])

        result = ops.concatenate(result, axis=self.channel_axis)
        return self.projection(result)

    def compute_output_shape(self, inputs_shape):
        if self.data_format == "channels_first":
            return tuple(
                (inputs_shape[0],) + (self.num_channels,) + (inputs_shape[2:])
            )
        else:
            return tuple((inputs_shape[:-1]) + (self.num_channels,))

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "dilation_rates": self.dilation_rates,
                "num_channels": self.num_channels,
                "activation": self.activation,
                "dropout": self.dropout,
            }
        )
        return config



import keras

from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.image_classifier import ImageClassifier
from keras_hub.src.models.task import Task
from keras_hub.src.models.vgg.vgg_backbone import VGGBackbone
from keras_hub.src.models.vgg.vgg_image_classifier_preprocessor import (
    VGGImageClassifierPreprocessor,
)


@keras_hub_export("keras_hub.models.VGGImageClassifier")
class VGGImageClassifier(ImageClassifier):
    """VGG image classification task.

    `VGGImageClassifier` tasks wrap a `keras_hub.models.VGGBackbone` and
    a `keras_hub.models.Preprocessor` to create a model that can be used for
    image classification. `VGGImageClassifier` tasks take an additional
    `num_classes` argument, controlling the number of predicted output classes.

    To fine-tune with `fit()`, pass a dataset containing tuples of `(x, y)`
    labels where `x` is a string and `y` is a integer from `[0, num_classes)`.

    Not that unlike `keras_hub.model.ImageClassifier`, the `VGGImageClassifier`
    allows and defaults to `pooling="flatten"`, when inputs are flatten and
    passed through two intermediate dense layers before the final output
    projection.

    Args:
        backbone: A `keras_hub.models.VGGBackbone` instance or a `keras.Model`.
        num_classes: int. The number of classes to predict.
        preprocessor: `None`, a `keras_hub.models.Preprocessor` instance,
            a `keras.Layer` instance, or a callable. If `None` no preprocessing
            will be applied to the inputs.
        pooling: `"flatten"`, `"avg"`, or `"max"`. The type of pooling to apply
            on backbone output. The default is flatten to match the original
            VGG implementation, where backbone inputs will be flattened and
            passed through two dense layers with a `"relu"` activation.
        pooling_hidden_dim: the output feature size of the pooling dense layers.
            This only applies when `pooling="flatten"`.
        activation: `None`, str, or callable. The activation function to use on
            the `Dense` layer. Set `activation=None` to return the output
            logits. Defaults to `"softmax"`.
        head_dtype: `None`, str, or `keras.mixed_precision.DTypePolicy`. The
            dtype to use for the classification head's computations and weights.


    Examples:

    Call `predict()` to run inference.
    ```python
    # Load preset and train
    images = np.random.randint(0, 256, size=(2, 224, 224, 3))
    classifier = keras_hub.models.VGGImageClassifier.from_preset(
        "vgg_16_imagenet"
    )
    classifier.predict(images)
    ```

    Call `fit()` on a single batch.
    ```python
    # Load preset and train
    images = np.random.randint(0, 256, size=(2, 224, 224, 3))
    labels = [0, 3]
    classifier = keras_hub.models.VGGImageClassifier.from_preset(
        "vgg_16_imagenet"
    )
    classifier.fit(x=images, y=labels, batch_size=2)
    ```

    Call `fit()` with custom loss, optimizer and backbone.
    ```python
    classifier = keras_hub.models.VGGImageClassifier.from_preset(
        "vgg_16_imagenet"
    )
    classifier.compile(
        loss=keras.losses.SparseCategoricalCrossentropy(from_logits=True),
        optimizer=keras.optimizers.Adam(5e-5),
    )
    classifier.backbone.trainable = False
    classifier.fit(x=images, y=labels, batch_size=2)
    ```

    Custom backbone.
    ```python
    images = np.random.randint(0, 256, size=(2, 224, 224, 3))
    labels = [0, 3]
    model = keras_hub.models.VGGBackbone(
        stackwise_num_repeats = [2, 2, 3, 3, 3],
        stackwise_num_filters = [64, 128, 256, 512, 512],
        image_shape = (224, 224, 3),
    )
    classifier = keras_hub.models.VGGImageClassifier(
        backbone=backbone,
        num_classes=4,
    )
    classifier.fit(x=images, y=labels, batch_size=2)
    ```
    """

    backbone_cls = VGGBackbone
    preprocessor_cls = VGGImageClassifierPreprocessor

    def __init__(
        self,
        backbone,
        num_classes,
        preprocessor=None,
        pooling="avg",
        pooling_hidden_dim=4096,
        activation=None,
        dropout=0.0,
        head_dtype=None,
        **kwargs,
    ):
        head_dtype = head_dtype or backbone.dtype_policy
        data_format = getattr(backbone, "data_format", None)

        # === Layers ===
        self.backbone = backbone
        self.preprocessor = preprocessor
        if pooling == "avg":
            self.pooler = keras.layers.GlobalAveragePooling2D(
                data_format,
                dtype=head_dtype,
                name="pooler",
            )
        elif pooling == "max":
            self.pooler = keras.layers.GlobalMaxPooling2D(
                data_format,
                dtype=head_dtype,
                name="pooler",
            )
        elif pooling == "flatten":
            self.pooler = keras.Sequential(
                [
                    keras.layers.Flatten(name="flatten"),
                    keras.layers.Dense(pooling_hidden_dim, activation="relu"),
                    keras.layers.Dense(pooling_hidden_dim, activation="relu"),
                ],
                name="pooler",
            )
        else:
            raise ValueError(
                "Unknown `pooling` type. Polling should be either `'avg'` or "
                f"`'max'`. Received: pooling={pooling}."
            )

        self.head = keras.Sequential(
            [
                keras.layers.Conv2D(
                    filters=4096,
                    kernel_size=7,
                    name="fc1",
                    activation=activation,
                    use_bias=True,
                    padding="same",
                ),
                keras.layers.Dropout(
                    rate=dropout,
                    dtype=head_dtype,
                    name="output_dropout",
                ),
                keras.layers.Conv2D(
                    filters=4096,
                    kernel_size=1,
                    name="fc2",
                    activation=activation,
                    use_bias=True,
                    padding="same",
                ),
                self.pooler,
                keras.layers.Dense(
                    num_classes,
                    activation=activation,
                    dtype=head_dtype,
                    name="predictions",
                ),
            ],
            name="head",
        )

        # === Functional Model ===
        inputs = self.backbone.input
        x = self.backbone(inputs)
        outputs = self.head(x)

        # Skip the parent class functional model.
        Task.__init__(
            self,
            inputs=inputs,
            outputs=outputs,
            **kwargs,
        )

        # === Config ===
        self.num_classes = num_classes
        self.activation = activation
        self.pooling = pooling
        self.pooling_hidden_dim = pooling_hidden_dim
        self.dropout = dropout
        self.preprocessor = preprocessor

    def get_config(self):
        # Backbone serialized in `super`
        config = super().get_config()
        config.update(
            {
                "num_classes": self.num_classes,
                "pooling": self.pooling,
                "activation": self.activation,
                "pooling_hidden_dim": self.pooling_hidden_dim,
                "dropout": self.dropout,
            }
        )
        return config



from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.csp_darknet.csp_darknet_backbone import (
    CSPDarkNetBackbone,
)
from keras_hub.src.models.image_classifier import ImageClassifier


@keras_hub_export("keras_hub.models.CSPDarkNetImageClassifier")
class CSPDarkNetImageClassifier(ImageClassifier):
    backbone_cls = CSPDarkNetBackbone



import sys

import keras
from absl import logging
from packaging.version import parse

try:
    import tensorflow as tf
except ImportError:
    tf = None


def clone_initializer(initializer):
    """Clones an initializer to ensure a new seed.

    As of tensorflow 2.10, we need to clone user passed initializers when
    invoking them twice to avoid creating the same randomized initialization.
    """
    # If we get a string or dict, just return as we cannot and should not clone.
    if not isinstance(initializer, keras.initializers.Initializer):
        return initializer
    config = initializer.get_config()
    return initializer.__class__.from_config(config)


def print_msg(message, line_break=True):
    """Print the message to absl logging or stdout."""
    # Copied from core Keras.
    if keras.utils.is_interactive_logging_enabled():
        if line_break:
            sys.stdout.write(message + "\n")
        else:
            sys.stdout.write(message)
        sys.stdout.flush()
    else:
        logging.info(message)


@keras.saving.register_keras_serializable(package="keras_hub")
def gelu_approximate(x):
    return keras.activations.gelu(x, approximate=True)


def has_quantization_support():
    return False if parse(keras.version()) < parse("3.4.0") else True


def assert_quantization_support():
    if not has_quantization_support():
        raise ValueError(
            "Quantization API requires Keras >= 3.4.0 to function "
            f"correctly. Received: '{keras.version()}'"
        )


def standardize_data_format(data_format):
    if data_format is None:
        return keras.config.image_data_format()
    data_format = str(data_format).lower()
    if data_format not in {"channels_first", "channels_last"}:
        raise ValueError(
            "The `data_format` argument must be one of "
            "{'channels_first', 'channels_last'}. "
            f"Received: data_format={data_format}"
        )
    return data_format



"""DenseNet preset configurations."""

backbone_presets = {
    "densenet_121_imagenet": {
        "metadata": {
            "description": (
                "121-layer DenseNet model pre-trained on the ImageNet 1k dataset "
                "at a 224x224 resolution."
            ),
            "params": 7037504,
            "official_name": "DenseNet",
            "path": "densenet",
            "model_card": "https://arxiv.org/abs/1608.06993",
        },
        "kaggle_handle": "kaggle://keras/densenet/keras/densenet_121_imagenet/2",
    },
    "densenet_169_imagenet": {
        "metadata": {
            "description": (
                "169-layer DenseNet model pre-trained on the ImageNet 1k dataset "
                "at a 224x224 resolution."
            ),
            "params": 12642880,
            "official_name": "DenseNet",
            "path": "densenet",
            "model_card": "https://arxiv.org/abs/1608.06993",
        },
        "kaggle_handle": "kaggle://keras/densenet/keras/densenet_169_imagenet/2",
    },
    "densenet_201_imagenet": {
        "metadata": {
            "description": (
                "201-layer DenseNet model pre-trained on the ImageNet 1k dataset "
                "at a 224x224 resolution."
            ),
            "params": 18321984,
            "official_name": "DenseNet",
            "path": "densenet",
            "model_card": "https://arxiv.org/abs/1608.06993",
        },
        "kaggle_handle": "kaggle://keras/densenet/keras/densenet_201_imagenet/2",
    },
}
